import ctypes
import math
import multiprocessing
import random
import subprocess
import time
import uuid
from pathlib import Path
from unittest.mock import Mock

import pytest
import sqlalchemy.dialects
import sqlalchemy.exc
from _pytest.monkeypatch import MonkeyPatch
from _pytest.tmpdir import TempdirFactory
from alembic.runtime.migration import MigrationContext
from sqlalchemy.orm import Session

import rasa.utils.io as rasa_io_utils
import rasax.community.database.utils as db_utils
import rasax.community.sql_migrations as migrations
from rasax.community import telemetry
from rasax.community import utils, config
from rasax.community.database.admin import Role
from tests.unit.conftest import insert_test_user

_PROCESS_TEST_TELEMETRY_ID = uuid.uuid4().hex


@pytest.fixture
def all_versions():
    return [
        "0.21.0a3.dev45+geda689e",
        "0.21.0a3.dev8+gb7f2938",
        "0.21.0a2",
        "0.21.0a2.dev25+gc41e902",
        "0.21.0a2.dev2+gda67ff5",
        "0.21.0a1",
        "0.20.3.dev157+g02d6625",
        "0.20.3.dev1+g2d271d2",
        "0.20.2",
        "0.20.2a3.dev158+gd656601",
        "0.20.2a3.dev47+g8ab683f",
        "0.20.2a2",
        "0.20.2a2.dev60+gf23b110",
        "0.20.2a2.dev3+gf451a26",
        "0.20.2a1",
        "0.20.2.dev569+g1665f89",
        "0.20.2.dev26+gae2d791",
        "0.20.1",
        "0.20.1.dev157+g92c8166",
        "0.20.1.dev21+g26647d5",
        "0.20.0",
        "0.20.0a1",
        "0.19.6.dev336+gc771b11",
    ]


@pytest.mark.parametrize(
    "current_version, newer_stable_version",
    [
        ("0.21.0a4.dev22+nwow", "0.21.0a4.dev22+nwow"),
        ("0.19.6.dev336+gc771b11", "0.20.2"),
        ("0.20.3.dev1+g2d271d2", "0.20.3.dev1+g2d271d2"),
        ("0.20.2", "0.20.2"),
        ("0.21.0a3.dev45+geda689e", "0.21.0a3.dev45+geda689e"),
        ("0.20.1", "0.20.2"),
    ],
)
def test_get_newer_stable_version(all_versions, current_version, newer_stable_version):
    assert (
        utils._get_newer_stable_version(all_versions, current_version)
        == newer_stable_version
    )


def test_query_result_to_dict():
    query_result = ["zero", "one", "two", "three"]
    fields_query = [("d", True), ("a.c", True), ("a.b", True), ("b.c.d", True)]

    actual = utils.query_result_to_dict(query_result, fields_query)

    assert actual["d"] == query_result[0]
    assert actual["a"]["c"] == query_result[1]
    assert actual["a"]["b"] == query_result[2]
    assert actual["b"]["c"]["d"] == query_result[3]


def test_foreign_keys_constraints(session: Session):
    insert_test_user(session, "some user")

    db_connection = session.connection()
    raw_connection = db_connection.connection.connection

    db_utils.set_sqlite_pragmas(raw_connection, enforce_foreign_keys=True)
    with pytest.raises(sqlalchemy.exc.DatabaseError):
        db_connection.execute(f"DROP table {Role.__tablename__};")


def test_is_db_revision_latest(session: Session):
    context = MigrationContext.configure(session.connection())
    heads = context.get_current_heads()
    current = context.get_current_revision()

    assert heads[0] == current
    assert db_utils.is_db_revision_latest(session)


def test_is_db_revision_latest_false(session_without_migration: Session):
    context = MigrationContext.configure(session_without_migration.connection())
    heads = context.get_current_heads()
    current = context.get_current_revision()

    assert not heads or heads[0] != current
    assert not db_utils.is_db_revision_latest(session_without_migration)


def test_turn_off_foreign_keys_constraints(session_without_migration: Session):
    migrations.run_migrations(session_without_migration)
    insert_test_user(session_without_migration, "some user")

    db_connection = session_without_migration.connection()
    raw_connection = db_connection.connection.connection

    db_utils.set_sqlite_pragmas(raw_connection, enforce_foreign_keys=False)
    if isinstance(db_connection.dialect, sqlalchemy.dialects.sqlite.dialect):
        db_connection.execute(f"DROP table {Role.__tablename__};")


def test_turn_off_then_on_foreign_key_constraints(session_without_migration: Session):
    migrations.run_migrations(session_without_migration)
    insert_test_user(session_without_migration, "some user")

    db_connection = session_without_migration.connection()
    raw_connection = db_connection.connection.connection

    db_utils.set_sqlite_pragmas(raw_connection, enforce_foreign_keys=False)
    db_utils.set_sqlite_pragmas(raw_connection, enforce_foreign_keys=True)

    with pytest.raises(sqlalchemy.exc.DatabaseError):
        db_connection.execute(f"DROP table {Role.__tablename__};")


@pytest.mark.parametrize(
    "before_truncation, expected",
    [(1.0, 1.0), (1.1234, 1.12), (19231231.00, 19231231.00)],
)
def test_truncate_float(before_truncation: float, expected: float):
    assert utils.truncate_float(before_truncation, decimal_places=2) == expected


def test_add_plural_suffix():
    base_string = "found item"
    base_string_with_replacement_field = base_string + "{}"

    # list with single item yields singular case
    assert (
        utils.add_plural_suffix(base_string_with_replacement_field, [1]) == "found item"
    )

    # list with multiple items yields plural case
    assert (
        utils.add_plural_suffix(base_string_with_replacement_field, [1, 2, 3])
        == "found items"
    )


def test_query_result_len():
    count = random.randint(1, 10)
    result = utils.QueryResult([{} for _ in range(count)], count)
    assert len(result) == count


def test_decode_base_64():
    import base64

    expected = "hello"
    encoded = base64.b64encode(expected.encode("utf-8")).decode("utf-8")

    assert utils.decode_base64(encoded) == expected


@pytest.mark.parametrize(
    "original, encoded", [("testing", "dGVzdGluZw=="), ("hello", "aGVsbG8=")]
)
def test_encode_base_64(original, encoded):
    assert utils.encode_base64(original) == encoded


def test_get_project_directory_if_none_is_given() -> None:
    utils.set_project_directory("")

    assert utils.get_project_directory() == Path()


def test_set_project_directory() -> None:
    directory = Path("test-directory")
    utils.set_project_directory(directory)

    assert utils.get_project_directory() == directory


def test_should_dump() -> None:
    with config.PROJECT_DIRECTORY.get_lock():
        config.PROJECT_DIRECTORY.value = b""

    assert not utils.should_dump()

    utils.set_project_directory("/tmp")

    assert utils.should_dump()


def test_write_text_to_file(tmpdir: Path) -> None:
    file_path = tmpdir / "sub" / "file.txt"

    assert not file_path.exists()

    utils.write_text_to_file(file_path, "some content")

    assert file_path.exists()


def test_dump_yaml_to_temporary_file():
    # dump dictionary to temporary yaml file
    data = {"some": "data"}
    temp_path = utils.dump_as_yaml_to_temporary_file(data)

    # read temporary file and check contents
    assert rasa_io_utils.read_yaml_file(temp_path) == data


def test_extract_partial_endpoint_config(tmpdir_factory: TempdirFactory):
    endpoint_config = f"""
event_broker:
    type: pika
other_section:
    type: do_not_extract
"""
    import rasa.nlu.utils as rasa_nlu_utils

    temp_dir = tmpdir_factory.mktemp("endpoint-config-test")
    config_path = temp_dir / "endpoints.yml"
    rasa_nlu_utils.write_to_file(config_path, endpoint_config)

    # read config partially
    extracted_cfg = utils.extract_partial_endpoint_config(config_path, "event_broker")

    # make sure it contains the desired section
    assert extracted_cfg.type == "pika"


def test_rasa_version_compatibility():
    # If Rasa and Rasa X versions for the same dependency conflict, this fails.
    assert subprocess.call(["rasa", "--help"]) == 0


def test_run_in_process(monkeypatch: MonkeyPatch):
    import multiprocessing

    # Use "fork" context for this test instead of global one
    # https://github.com/pytest-dev/pytest/issues/958
    monkeypatch.setattr(
        utils, "mp_context", lambda: multiprocessing.get_context("fork")
    )

    import time

    context = utils.mp_context()
    val = context.Value(ctypes.c_int, 0)
    assert val.value == 0

    def _run_in_process_helper(arg: multiprocessing.Value) -> None:
        arg.value = 42

    utils.run_in_process(_run_in_process_helper, args=(val,))
    time.sleep(0.5)

    assert val.value == 42


def test_uptime():
    start = utils.get_uptime()
    assert start > 0
    time.sleep(1)

    end = utils.get_uptime()
    assert math.isclose(end - start, 1.0, abs_tol=0.1)


def _run_in_process_telemetry_config_helper() -> None:
    telemetry._telemetry_enabled.value = True
    telemetry._telemetry_id.value = _PROCESS_TEST_TELEMETRY_ID.encode(
        utils.DEFAULT_ENCODING
    )


def test_run_in_process_telemetry_config(monkeypatch: MonkeyPatch):
    import time

    monkeypatch.setattr(
        utils, "mp_context", lambda: multiprocessing.get_context("fork")
    )

    monkeypatch.setattr(
        telemetry, "_telemetry_enabled", utils.mp_context().Value(ctypes.c_bool, False)
    )
    monkeypatch.setattr(
        telemetry,
        "_telemetry_id",
        utils.mp_context().Array(ctypes.c_char, telemetry.TELEMETRY_ID_LENGTH),
    )

    assert not telemetry._telemetry_enabled.value
    assert telemetry._telemetry_id.value == b""

    utils.run_in_process(_run_in_process_telemetry_config_helper, daemon=True)
    time.sleep(0.5)

    # Because the telemetry configuration is shared among processes, changing its values
    # in a child process should affect the parent process:
    assert telemetry._telemetry_enabled.value
    assert telemetry._telemetry_id.value == _PROCESS_TEST_TELEMETRY_ID.encode(
        utils.DEFAULT_ENCODING
    )


def test_run_in_process_telemetry_config_uninitialized(monkeypatch: MonkeyPatch):
    import time

    monkeypatch.setattr(
        utils, "mp_context", lambda: multiprocessing.get_context("fork")
    )

    # Same test as above, but this time the parent process hasn't initialized
    # its telemetry config
    monkeypatch.setattr(telemetry, "_telemetry_enabled", None)
    monkeypatch.setattr(telemetry, "_telemetry_id", None)

    utils.run_in_process(_run_in_process_telemetry_config_helper, daemon=True)
    time.sleep(0.5)

    # Values should be now initialized (set to Array/Value), and have values
    # inside
    assert telemetry._telemetry_enabled.value
    assert telemetry._telemetry_id.value == _PROCESS_TEST_TELEMETRY_ID.encode(
        utils.DEFAULT_ENCODING
    )

    def test_query_request_parameter_list_arg_empty():
        parameter_name = "TEST_PARAMETER"

        request = Mock()
        request.args = {}

        assert not utils.list_arg(request, parameter_name)

    def test_query_request_parameter_list():
        parameter_name = "TEST_PARAMETER"

        request = Mock()
        request.args = {parameter_name: "1,2,3,4", "other": "1"}

        assert len(utils.list_arg(request, parameter_name)) == 4
